@@ -1,9 +1,9 @@
 import abc
 import warnings
 from fractions import Fraction
-from types import MappingProxyType
-from typing import (Any, Callable, Hashable, List, Mapping, MutableMapping,
-                    NamedTuple, Optional, Union, cast)
+from types import MappingProxyType, SimpleNamespace
+from typing import (Any, Callable, ChainMap, Hashable, List, Mapping,
+                    MutableMapping, NamedTuple, Optional, Union, cast)
 
 import numpy as np
 import pandas
@@ -80,6 +80,28 @@ def convert_list(
         return [conversion_function(r_list.value[0]), *cdr]
 
 
+def convert_env(
+    r_env: parser.RObject,
+    conversion_function: Callable[
+        [Union[parser.RData, parser.RObject]
+         ], Any]=lambda x: x
+) -> ChainMap[Union[str, bytes], Any]:
+
+    if r_env.info.type is not parser.RObjectType.ENV:
+        raise TypeError("Must receive a ENV object")
+
+    frame = conversion_function(r_env.value.frame)
+    enclosure = conversion_function(r_env.value.enclosure)
+    hash_table = conversion_function(r_env.value.hash_table)
+
+    dictionary = {}
+    for d in hash_table:
+        if d is not None:
+            dictionary.update(d)
+
+    return ChainMap(dictionary, enclosure)
+
+
 def convert_attrs(
     r_obj: parser.RObject,
         conversion_function: Callable[
@@ -169,7 +191,24 @@ def convert_vector(
     return value
 
 
-def convert_char(r_char: parser.RObject) -> Union[str, bytes]:
+def safe_decode(byte_str: bytes, encoding: str) -> Union[str, bytes]:
+    """
+    Decode a (possibly malformed) string.
+    """
+    try:
+        return byte_str.decode(encoding)
+    except UnicodeDecodeError as e:
+        warnings.warn(
+            f"Exception while decoding {byte_str!r}: {e}",
+        )
+        return byte_str
+
+
+def convert_char(
+    r_char: parser.RObject,
+    default_encoding: Optional[str] = None,
+    force_default_encoding: bool = False,
+) -> Union[str, bytes, None]:
     """
     Decode a R character array to a Python string or bytes.
 
@@ -195,20 +234,27 @@ def convert_char(r_char: parser.RObject) -> Union[str, bytes]:
     if r_char.info.type is not parser.RObjectType.CHAR:
         raise TypeError("Must receive a CHAR object")
 
+    if r_char.value is None:
+        return None
+
     assert isinstance(r_char.value, bytes)
 
-    if r_char.info.gp & parser.CharFlags.UTF8:
-        return r_char.value.decode("utf_8")
-    elif r_char.info.gp & parser.CharFlags.LATIN1:
-        return r_char.value.decode("latin_1")
-    elif r_char.info.gp & parser.CharFlags.ASCII:
-        return r_char.value.decode("ascii")
-    elif r_char.info.gp & parser.CharFlags.BYTES:
-        return r_char.value
+    if not force_default_encoding:
+        if r_char.info.gp & parser.CharFlags.UTF8:
+            return safe_decode(r_char.value, "utf_8")
+        elif r_char.info.gp & parser.CharFlags.LATIN1:
+            return safe_decode(r_char.value, "latin_1")
+        elif r_char.info.gp & parser.CharFlags.ASCII:
+            return safe_decode(r_char.value, "ascii")
+        elif r_char.info.gp & parser.CharFlags.BYTES:
+            return r_char.value
+
+    if default_encoding:
+        return safe_decode(r_char.value, default_encoding)
     else:
         # Assume ASCII if no encoding is marked
         warnings.warn(f"Unknown encoding. Assumed ASCII.")
-        return r_char.value.decode("ascii")
+        return safe_decode(r_char.value, "ascii")
 
 
 def convert_symbol(r_symbol: parser.RObject,
@@ -401,7 +447,7 @@ class SimpleConverter(Converter):
 
     Parameters
     ----------
-    constructor_dict: dict
+    constructor_dict:
         Dictionary mapping names of R classes to constructor functions with
         the following prototype:
 
@@ -413,20 +459,39 @@ class SimpleConverter(Converter):
         the dictionary used is
         :data:`~rdata.conversion._conversion.DEFAULT_CLASS_MAP`
         which has support for several common classes.
+    default_encoding:
+        Default encoding used for strings with unknown encoding. If `None`,
+        the one stored in the file will be used, or ASCII as a fallback.
+    force_default_encoding;
+        Use the default encoding even if the strings specify other encoding.
 
     """
 
-    def __init__(self,
-                 constructor_dict: Mapping[
-                     Union[str, bytes],
-                     Constructor]=DEFAULT_CLASS_MAP) -> None:
+    def __init__(
+        self,
+        constructor_dict: Mapping[
+            Union[str, bytes],
+            Constructor,
+        ] = DEFAULT_CLASS_MAP,
+        default_encoding: Optional[str] = None,
+        force_default_encoding: bool = False,
+        global_environment: Optional[Mapping[Union[str, bytes], Any]] = None,
+    ) -> None:
 
         self.constructor_dict = constructor_dict
+        self.default_encoding = default_encoding
+        self.force_default_encoding = force_default_encoding
+        self.global_environment = ChainMap(
+            {} if global_environment is None
+            else global_environment
+        )
+        self.empty_environment = ChainMap({})
 
         self._reset()
 
     def _reset(self) -> None:
         self.references: MutableMapping[int, Any] = {}
+        self.default_encoding_used = self.default_encoding
 
     def convert(self, data: Union[parser.RData, parser.RObject]) -> Any:
         self._reset()
@@ -440,6 +505,8 @@ class SimpleConverter(Converter):
         obj: RObject
         if isinstance(data, parser.RData):
             obj = data.object
+            if self.default_encoding is None:
+                self.default_encoding_used = data.extra.encoding
         else:
             obj = data
 
@@ -462,6 +529,11 @@ class SimpleConverter(Converter):
             # Expand the list and process the elements
             value = convert_list(obj, self._convert_next)
 
+        elif obj.info.type == parser.RObjectType.ENV:
+
+            # Return a ChainMap of the environments
+            value = convert_env(obj, self._convert_next)
+
         elif obj.info.type == parser.RObjectType.LANG:
 
             # Expand the list and process the elements, returning a
@@ -474,7 +546,11 @@ class SimpleConverter(Converter):
         elif obj.info.type == parser.RObjectType.CHAR:
 
             # Return the internal string
-            value = convert_char(obj)
+            value = convert_char(
+                obj,
+                default_encoding=self.default_encoding_used,
+                force_default_encoding=self.force_default_encoding,
+            )
 
         elif obj.info.type in {parser.RObjectType.LGL,
                                parser.RObjectType.INT,
@@ -502,6 +578,15 @@ class SimpleConverter(Converter):
             # Convert the internal objects returning a special object
             value = RExpression(rexpression_list)
 
+        elif obj.info.type == parser.RObjectType.S4:
+            value = SimpleNamespace(**attrs)
+
+        elif obj.info.type == parser.RObjectType.EMPTYENV:
+            value = self.empty_environment
+
+        elif obj.info.type == parser.RObjectType.GLOBALENV:
+            value = self.global_environment
+
         elif obj.info.type == parser.RObjectType.REF:
 
             # Return the referenced value
