@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import abc
 import bz2
 import enum
@@ -7,8 +9,8 @@ import os
 import pathlib
 import warnings
 import xdrlib
-from typing import (Any, BinaryIO, List, NamedTuple, Optional, Set, TextIO,
-                    Union)
+from dataclasses import dataclass
+from typing import Any, BinaryIO, List, Optional, Set, TextIO, Union
 
 import numpy as np
 
@@ -20,14 +22,16 @@ class FileTypes(enum.Enum):
     bzip2 = "bz2"
     gzip = "gzip"
     xz = "xz"
-    rdata_binary = "rdata (binary)"
+    rdata_binary_v2 = "rdata version 2 (binary)"
+    rdata_binary_v3 = "rdata version 3 (binary)"
 
 
 magic_dict = {
     FileTypes.bzip2: b"\x42\x5a\x68",
     FileTypes.gzip: b"\x1f\x8b",
     FileTypes.xz: b"\xFD7zXZ\x00",
-    FileTypes.rdata_binary: b"RDX2\n"
+    FileTypes.rdata_binary_v2: b"RDX2\n",
+    FileTypes.rdata_binary_v3: b"RDX3\n"
 }
 
 
@@ -97,6 +101,8 @@ class RObjectType(enum.Enum):
     WEAKREF = 23  # weak reference
     RAW = 24  # raw vector
     S4 = 25  # S4 classes not of simple type
+    EMPTYENV = 242  # Empty environment
+    GLOBALENV = 253  # Global environment
     NILVALUE = 254  # NIL value
     REF = 255  # Reference
 
@@ -110,7 +116,8 @@ class CharFlags(enum.IntFlag):
     ASCII = 1 << 6
 
 
-class RVersions(NamedTuple):
+@dataclass
+class RVersions():
     """
     R versions.
     """
@@ -119,7 +126,18 @@ class RVersions(NamedTuple):
     minimum: int
 
 
-class RObjectInfo(NamedTuple):
+@dataclass
+class RExtraInfo():
+    """
+    Extra information.
+
+    Contains the default encoding (only in version 3).
+    """
+    encoding: Optional[str] = None
+
+
+@dataclass
+class RObjectInfo():
     """
     Internal attributes of a R object.
     """
@@ -131,15 +149,16 @@ class RObjectInfo(NamedTuple):
     reference: int
 
 
-class RObject(NamedTuple):
+@dataclass
+class RObject():
     """
     Representation of a R object.
     """
     info: RObjectInfo
     value: Any
-    attributes: Optional['RObject']  # type: ignore
-    tag: Optional['RObject'] = None  # type: ignore
-    referenced_object: Optional['RObject'] = None  # type: ignore
+    attributes: Optional[RObject]
+    tag: Optional[RObject] = None
+    referenced_object: Optional[RObject] = None
 
     def _str_internal(
         self,
@@ -199,14 +218,27 @@ class RObject(NamedTuple):
         return self._str_internal()
 
 
-class RData(NamedTuple):
+@dataclass
+class RData():
     """
     Data contained in a R file.
     """
     versions: RVersions
+    extra: RExtraInfo
     object: RObject
 
 
+@dataclass
+class EnvironmentValue():
+    """
+    Value of an environment.
+    """
+    locked: bool
+    enclosure: RObject
+    frame: RObject
+    hash_table: RObject
+
+
 class Parser(abc.ABC):
     """
     Parser interface for a R file.
@@ -251,9 +283,10 @@ class Parser(abc.ABC):
         """
 
         versions = self.parse_versions()
+        extra_info = self.parse_extra_info(versions)
         obj = self.parse_R_object()
 
-        return RData(versions, obj)
+        return RData(versions, extra_info, obj)
 
     def parse_versions(self) -> RVersions:
         """
@@ -264,12 +297,28 @@ class Parser(abc.ABC):
         r_version = self.parse_int()
         minimum_r_version = self.parse_int()
 
-        if format_version != 2:
+        if format_version not in [2, 3]:
             raise NotImplementedError(
-                "Format version {format_version} unsupported")
+                f"Format version {format_version} unsupported",
+            )
 
         return RVersions(format_version, r_version, minimum_r_version)
 
+    def parse_extra_info(self, versions: RVersions) -> RExtraInfo:
+        """
+        Parse the versions header.
+        """
+
+        encoding = None
+
+        if versions.format >= 3:
+            encoding_len = self.parse_int()
+            encoding = self.parse_string(encoding_len).decode("ASCII")
+
+        extra_info = RExtraInfo(encoding)
+
+        return extra_info
+
     def parse_R_object(
         self,
         reference_list: Optional[List[RObject]] = None
@@ -294,9 +343,14 @@ class Parser(abc.ABC):
         attributes_read = False
         add_reference = False
 
+        result = None
+
         value: Any
 
-        if info.type == RObjectType.SYM:
+        if info.type == RObjectType.NIL:
+            value = None
+
+        elif info.type == RObjectType.SYM:
             # Read Char
             value = self.parse_R_object(reference_list)
             # Symbols can be referenced
@@ -315,12 +369,38 @@ class Parser(abc.ABC):
             cdr = self.parse_R_object(reference_list)
             value = (car, cdr)
 
+        elif info.type == RObjectType.ENV:
+            result = RObject(
+                info=info,
+                tag=tag,
+                attributes=attributes,
+                value=None,
+                referenced_object=referenced_object,
+            )
+
+            reference_list.append(result)
+
+            locked = self.parse_bool()
+            enclosure = self.parse_R_object(reference_list)
+            frame = self.parse_R_object(reference_list)
+            hash_table = self.parse_R_object(reference_list)
+            attributes = self.parse_R_object(reference_list)
+
+            value = EnvironmentValue(
+                locked=locked,
+                enclosure=enclosure,
+                frame=frame,
+                hash_table=hash_table,
+            )
+
         elif info.type == RObjectType.CHAR:
             length = self.parse_int()
             if length > 0:
                 value = self.parse_string(length=length)
             elif length == 0:
                 value = b""
+            elif length == -1:
+                value = None
             else:
                 raise NotImplementedError(
                     f"Length of CHAR cannot be {length}")
@@ -366,6 +446,15 @@ class Parser(abc.ABC):
             for i in range(length):
                 value[i] = self.parse_R_object(reference_list)
 
+        elif info.type == RObjectType.S4:
+            value = None
+
+        elif info.type == RObjectType.EMPTYENV:
+            value = None
+
+        elif info.type == RObjectType.GLOBALENV:
+            value = None
+
         elif info.type == RObjectType.NILVALUE:
             value = None
 
@@ -383,10 +472,19 @@ class Parser(abc.ABC):
         if info.attributes and not attributes_read:
             attributes = self.parse_R_object(reference_list)
 
-        result = RObject(info=info, tag=tag,
-                         attributes=attributes,
-                         value=value,
-                         referenced_object=referenced_object)
+        if result is None:
+            result = RObject(
+                info=info,
+                tag=tag,
+                attributes=attributes,
+                value=value,
+                referenced_object=referenced_object,
+            )
+        else:
+            result.info = info
+            result.attributes = attributes
+            result.value = value
+            result.referenced_object = referenced_object
 
         if add_reference:
             reference_list.append(result)
@@ -451,6 +549,7 @@ def parse_file(file_or_path: Union[BinaryIO, TextIO, 'os.PathLike[Any]',
         RData(versions=RVersions(format=2,
                                  serialized=196610,
                                  minimum=131840),
+              extra=RExtraInfo(encoding=None),
               object=RObject(info=RObjectInfo(type=<RObjectType.LIST: 2>,
                              object=False,
                              attributes=False,
@@ -541,6 +640,7 @@ def parse_data(data: bytes) -> RData:
         RData(versions=RVersions(format=2,
                                  serialized=196610,
                                  minimum=131840),
+              extra=RExtraInfo(encoding=None),
               object=RObject(info=RObjectInfo(type=<RObjectType.LIST: 2>,
                              object=False,
                              attributes=False,
@@ -600,7 +700,7 @@ def parse_data(data: bytes) -> RData:
         return parse_data(gzip.decompress(data))
     elif filetype is FileTypes.xz:
         return parse_data(lzma.decompress(data))
-    elif filetype is FileTypes.rdata_binary:
+    elif filetype in {FileTypes.rdata_binary_v2, FileTypes.rdata_binary_v3}:
         view = view[len(magic_dict[filetype]):]
         return parse_rdata_binary(view)
     else:
